
def ngrams(string, n = 3):
    '''take in string and create ngrams of length 3 for TF-IDF vectorization'''
    ngrams = zip(*[string[i:] for i in range(n)])
    return [''.join(ngram) for ngram in ngrams]


import pandas as pd
from loguru import logger


def merge_to_raw_df(raw_df: pd.DataFrame, mapping: dict) -> pd.DataFrame:
    """
    append the clean name to the original dataset and return the appended DataFrame.
    """

    dict_for_df = [{'first_plaintiff': k, 'first_plaintiff_clean': v} for k,v in mapping.items()]
    mapping_df = pd.DataFrame.from_dict(dict_for_df)
    merged_df = raw_df.merge(mapping_df, on=['first_plaintiff'])
    return merged_df 




def map_values(aStr: str, aDict: dict) -> str:
    """
    Look up if str in dict otherwise return ""
    """

    try:
        return aDict[aStr] 
    except KeyError:
        return ""



def combine_columns(clean,person,company):
    if(company == "" and person == ""):
        return clean
    
    if(company != "" and person != ""):
        logger.warning("NAME HAS BOTH COMPANY AND PERSON NAME: " + clean)
        return clean
    if(company != ""):
        return company

    if(person != ""):
        return person



def combine_columns_no_space(all_names, no_space):
    if(no_space == ""):
        return all_names
    
    return no_space


from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer
import re
import pandas as pd
from cleancourt.compute_ngrams import ngrams
from loguru import logger
from sklearn.neighbors import NearestNeighbors
import time
import numpy as np
from tqdm import tqdm
from cleancourt.integrate_output import map_values
import json
from cleancourt.filter_types import filter_companies




SIMILARITY_DISTANCE = .25   

def _print_matches(distances, indices, org_name_messy, org_name_clean):
    '''Print function to see output of matches'''
    
    all_matches = [(round(distances[i][0], 2), org_name_messy[i], org_name_clean[j[0]]) 
               for i, j in enumerate(indices)]
    all_matches = pd.DataFrame(all_matches, 
                           columns=['score', 'original', 'matched'])
    
    all_matches.to_csv('data/ignore_all_matches.csv')




def compare_companies(org_name_messy, org_name_clean, threshold = SIMILARITY_DISTANCE, print_all_matches = False):
    '''This method implements a K-Nearest Neighbors (KNN) algorithm to
    compare a standard list of *clean* party names to a *messy* list of
    party names. 

    With KNN algorithms, individual (new) data points are added to an existing
    dataset, and the "distance" to all other points in the dataset is computed.

    "Distance" is measured using TFIDF and cosine similarities similarly to the link_company_names_function. 

    '''

    # Remove people from the list of names
    org_name_messy = [name for name in org_name_messy if filter_companies(name)]


    logger.info("Vectorizing and Transforming Clean Names")

    # Create a sparse matrix of "ngrams" across the list of clean names.
    # Output of tfidf sparse matrix (and what you would see if you ran print(tfidf)) is below:
        # The tuple represents: (clean_name id, ngram id)
        # The value following the tuple represents the tf-idf score of a given ngram in a given clean_name
        # The tuples that are not there have a tf-idf score of 0 
    vectorizer = TfidfVectorizer(analyzer=ngrams, lowercase=True)
    tfidf = vectorizer.fit_transform(org_name_clean)
    


    logger.info("Computing Nearest Neighbor Algorithm")

    # This model uses the cosine similarity metric and the brute force metric
    # This algorithm was chosen to produce similar output based on other comparitors in this library.
    # However, in cases with extrememly large clean_names, this will be an inefficient method for computing 
    # as the brute force method computes the distances between all pairs of points in the dataset.

    # n_neighbors is set to 1 as we are only concerned with pulling the closest company name

    nbrs = NearestNeighbors(n_neighbors=1,
                            metric='cosine',
                            algorithm='brute',
                            n_jobs=-1).fit(tfidf)

    
    
    # Returns the distance to the next closest neighbor in the geographic plain, 
    # Also returns the index in the list of the closest neighbor in the clean data set
    distances, indices = nbrs.kneighbors(vectorizer.transform(org_name_messy))
    
    if(print_all_matches):
        _print_matches(distances, indices, org_name_messy, org_name_clean)

    logger.info("Pairing Names")

    # Creates a dict that pairs the messy_name with the closest clean name
    # Only adds the values to a dict if they are below a certain distance as determined by SIMILARITY_DISTANCE
    matches = {}
    for i in range(0, len(indices)):
        if(distances[i][0] < threshold):
            matches[org_name_messy[i]] = org_name_clean[indices[i][0]]

    return matches









from cleancourt.clean_data import clean_data
from cleancourt.check_spaces import check_spaces
from cleancourt.compare_companies import compare_companies
from cleancourt.cosine_similarity import link_company_names
from cleancourt.filter_types import filter_companies, filter_people
from cleancourt.integrate_output import *
from cleancourt.parse_names import format_people_names
from cleancourt.test_utils import generate_diagnostics
from cleancourt.full_clean_names import *
from cleancourt.separate_parent_company import separate_company_names



import ftfy
import re

def clean_data(string):
    '''Takes in a string and cleans it using regex and other standard functions'''

    if not isinstance(string, str): 
        string = ""

    string = ftfy.fix_text(string) # Fix text to ensure any non-ascii characters appear properly
    string = string.encode("ascii", errors="ignore").decode() # Remove these non-ascii characters
    string = string.lower() # Convert all characters to lowercase

    chars_to_remove = [")", "(", "[", "]", "{", "}", "'", "#", ";"] # remove these characters    
    rx = '[' + re.escape(''.join(chars_to_remove)) + ']'
    string = re.sub(rx, '', string)

    # Replace special characters and symbols
    string = string.replace('&', 'and') 
    string = string.replace('@', 'at')
    string = string.replace('h/w', '') # Sometimes appears next to party names that are husband (h) and wife (w)
    
    
    string = re.sub(' +', ' ', string).strip() # Remove multiple spaces and spaces at the end of a word

    string = re.sub("^the ", "", string) # Remove 'the' at the beginning of party names
    string = re.sub(' i+ ', ' ', string) # Remove one or more 'i' characters that appear in isolation (denotes numbers)

    
    

    string = re.sub(' c/o.*', '', string) # Remove care of


    string = re.sub(" corporation(\s|$)", " corp\\1", string) # shorten corporation to corp
    string = re.sub(" company(\s|$)", " co\\1", string) # shorten company to co
    string = re.sub(" maa?nage?ment(\s|$)", " mgmt\\1", string)  # etc
    string = re.sub(" incorporated(\s|$)", " inc\\1", string)
    string = re.sub(" apartments(\s|$)", " apts\\1", string)
    string = re.sub( " square(\s|$)", " sq\\1", string)
   
    

    string = re.sub(" et\.?\s?al\.?$", "", string)
    string = re.sub( "p\.?c\.?$", "", string) # Remove P.C. (Personal Corporation) abbreviation on the end of a name
    

    string = re.sub( "\.", "", string) # Remove any periods

    return(string)



import probablepeople as pp
import re

def filter_companies(name):

    '''Take in a particular string and determine whether or not the string is a person using the probablepeople library
    Assert that 
        1) ProbablePeople doesn't tag the name as a person OR
        2) the substring `|and|` is found in the name, which is much easier to parse as a company OR
        3) The name contains no spaces, which ProbablePeople likes to parse as a person AND
        4) The name does not start with a string, which these methods are poor at parsing right now'''


    # IMPORTANT: Cosine similarity (and really any fuzzy string match) doesn't work well on people
    # Think Smith, Tom A. and Smith, Tom B. -> are these people the same? A fuzzy string match would say yes. But can we be sure?
    # As such, for this its better to remove people from the output all together and deal with them later in the normalization process.
    # Fortunately the probablepeople library can be used to quickly identify strings that are people, and just ignore them. 

    # Probable people performs poorly on bank names. Working to update training dataset but this is a quick fix for now

    try:
        return ((pp.tag(name)[1] != 'Person') or ("|and|" in name) or (name.count(" ") == 0) or (re.search(' bank$', name)) or ("bennington crossing" in name)) and not name[0].isdigit() 
    except:
        return True



def filter_people(name):
    '''Take in a particular string and determine whether or not the string is a person using the probablepeople library
    Assert that 
        1) ProbablePeople tags the name as a person AND
        2) the substring `|and|` is not found in the name, which is much easier to parse as a company AND
        3) The name contains spaces (mononyms are often read as people by probablepeople. But more frequently refer to company names)
    '''

    try:
        val = pp.tag(name)
        return (val[1] == 'Person') and ("|and|" not in name) and (name.count(" ") != 0) and not (re.search(' bank$', name)) and ("bennington crossing" not in name)
    except:
        return False


from setuptools import setup, find_packages

setup(

    name='cleancourt',
    version='0.4.5',
    description='a library for cleaning court docket entries',
    author='Logan Pratico',
    author_email='praticol@lsc.gov',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        ],
    keywords="name standardization, plaintiff names, court data, name cleaning",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.9, <4",
    install_requires=["sklearn", "loguru", "probablepeople", "rapidfuzz", "tqdm", "pandas", "sparse-dot-topn", "ftfy"],
    project_urls={
            

        },

        )
